/*****************************************************************
Counter Firmware for Geiger Mueller counter. Uses 8x2 Char LCD
to display output. Timer0 and tmr0_msb variable are used to count
events during a one second interval, which are stored in counter.
The number of counts per second are stored in an array, later summed
and multiplied by 6 to get the CPM.

PD3 (INT1) is used to change between counting modes. Counting modes 
include per minute and accumulated. Counting input is PD4 (T0).

Data is sent over the serial port using a master/slave protocol.
No data is sent until requested by the master. Count values are
sent as soon as they are ready, but only once cleared to do so by
the master. Specifications of the GM tube are hard-coded into the
firmware.

Update interval is 1000ms.
Currently used with a "GM16/60B" Geiger tube. Conversion rate and
highest count rate are unknown. Max count rate set to 1500, and
dose rate to 357,14. This is based on the assumption that background
radiation where I am testing is 0.14Sv/hour, and the tube gives
roughly 50 clicks per minute on average.


Project: GM Counter
MCU Type: ATtiny2313
Clock: 8,00MHz

Data Types:

int8_t - Signed Char 
uint16_t - Unsigned Int 
uint32_t - Unsigned Long 
int64_t - Signed Long Long

(signed/unsigned) char - 1 byte 
(signed/unsigned) short - 2 bytes 
(signed/unsigned) int - 2 bytes 
(signed/unsigned) long - 4 bytes 
(signed/unsigned) long long - 8 bytes 
float - 4 bytes (floating point) 
double - alias to float

******************************************************************/

// Flag definitions
#define READY 0
#define TX_ENABLED 1
#define RX_TRIGGERED 2

// Communication protocol definitions
// To GM counter
#define REQUEST_GM_TUBE_DATA "READC"
#define START_GM "START"
#define STOP_GM "HALTT"
// From MCU
#define HEADER_TIME "PERID:1000"
#define HEADER_MAXCOUNT "MAXCT:1500"
#define HEADER_NAME "NAMET:GM16/60B"
#define HEADER_DOSERATE "DOSER:357.14"
#define DATAGRAM_COUNT "COUNT:"

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <avr/pgmspace.h>
#include "lcd.h"
#include "usart_lib.h"

// Declare global variables
volatile uint8_t tmr0_msb;             			// increments with every timer 0 overflow
volatile uint16_t counter;             			// Total number of events captured in 1 second
volatile uint8_t flag_reg;             			// Register for flags used in program

char stringBuffer[32];							// String buffer for copying out strings from prog mem

// Serial protocol constants
// Data sent from the MCU
const char DATAGRAM_STRING[] PROGMEM = DATAGRAM_COUNT;
const char NAME_STRING[] PROGMEM = HEADER_NAME;
const char TIME_STRING[] PROGMEM = HEADER_TIME;
const char MAXCOUNT_STRING[] PROGMEM = HEADER_MAXCOUNT;
const char DOSERATE_STRING[] PROGMEM = HEADER_DOSERATE;
// Data received by the MCU
/*const char REQUEST_CONFIG_STRING[] PROGMEM = REQUEST_GM_TUBE_DATA;
const char START_STRING[] PROGMEM = START_GM;
const char STOP_STRING[] PROGMEM = STOP_GM;*/

/*PGM_P const recv_protocol_string_table[] PROGMEM =
{
	REQUEST_CONFIG_STRING,
	START_STRING,
	STOP_STRING
};*/

PGM_P const send_protocol_string_table[] PROGMEM =
{
	NAME_STRING,
	TIME_STRING,
	MAXCOUNT_STRING,
	DOSERATE_STRING,
	DATAGRAM_STRING
};

// Find Counts per minute
uint16_t count_math(uint16_t newCount);
// Write value to LCD
void displayNumberOnLCDLine(uint16_t data, uint8_t line);
// Parse and send a counts value
void sendCountValue(uint16_t count);
// Send the stored configuration over serial
void sendMyConfigurationData(void);
// Parse a character command
void parseCommandChar(char command);

// USART Received Data Complete interrupt service routine
ISR(USART_RX_vect) {
	// Place the USART contents in the buffer, so it can be handled later
	bufferInsert(UDR);
}

// Timer 0 overflow interrupt service routine
ISR(TIMER0_OVF_vect)
{
    tmr0_msb++;
}

// Timer 1 overflow interrupt service routine
ISR(TIMER1_OVF_vect)
{
	// Disable Interrupts
	cli();
    
	// Find total number of events during last second
    counter = (255 * tmr0_msb) + TCNT0;
	
	flag_reg |= (1 << READY);
	
	// Set Timer 1 start point for 1 second interrupt rate
    TCNT1H = 0x7A;
	TCNT1L = 0x11;
	
	// Reinitialize Timer values, Enable Interrupts
	tmr0_msb = 0;
    TCNT0 = 0;
	sei();

}

// External Interrupt Routine, connected to front panel button
ISR(INT1_vect)
{	
	// Toggle transmission enable
	flag_reg ^= (1 << TX_ENABLED);
}

// Main
int main(void)
{
	uint16_t cpmValue;
	char latestReceivedByte;

	// Set PORTD pull-ups disabled
	PORTD = 0;

	// Timer/Counter 0 initialization, set clock source to T0 pin rising Edge
	TCCR0B = (1<<CS02)|(1<<CS01)|(1<<CS00);
	counter = 0;
	
	// Timer/Counter 1 initialization
	// Set Prescaler to 1:256, and compare match every 31250 cycles, for 1000ms interrupt rate. This means timer compare must happen at 31249 (0x7A11)
	TCCR1B = (1<<CS12)|(0<<CS11)|(0<<CS10);
	TCNT1H = 0x7A;
	TCNT1L = 0x11;
	
	// Timer 0 and 1 Overflow Interrupt initialization
	TIMSK = (1<<TOIE1)|(1<<TOIE0);

	// Turn off analog comparator to save power
	ACSR = (1<<ACD);

	// INT1 External Interrupt enable
	GIMSK = (1<<INT1);
	// ...on INT1 falling edge
	MCUCR = (1<<ISC11)|(0<<ISC10);

	// Initialize UART
	USART_Init();

	// LCD module initialization
	lcd_init(LCD_DISP_ON);
	lcd_clrscr();

	// Set default mode
	flag_reg &= ~(1 << TX_ENABLED);
	flag_reg |= (1 << RX_TRIGGERED);	// Ready to act on the first byte received

	sei();

	while(1)
	{
	
		// New byte received
		if (writeIndex != currentIndex) {
			
			latestReceivedByte = popFromBuffer();
			
			// Once a LF char is received, the next character should be the start of a new command
			// No other data is needed. Check if the last char was a LF, if so use the data.
			// Otherwise discard it.
			if (bit_is_set(flag_reg, RX_TRIGGERED)) {
				// Act on the command recevied
				parseCommandChar(latestReceivedByte);
				
				flag_reg &= ~(1 << RX_TRIGGERED);
				
			}else if (latestReceivedByte == '\n') {
				// Ready to act on the next byte received
				flag_reg |= (1 << RX_TRIGGERED);
				
			}
			
		}
		
		
		// Simple check so display in only updated once per second
		if (bit_is_set(flag_reg, READY))
		{
			// New count data, display and send out via UART
			if (bit_is_set(flag_reg, TX_ENABLED)) {
				sendCountValue(counter);
				
			}
			
			// Calculate the counts per minute
			cpmValue = count_math(counter);
			
			// Display counts per minute
			lcd_clrscr();
			lcd_puts("cnt/min");
			displayNumberOnLCDLine(cpmValue, 1);
			
			// Clear ready flag
			flag_reg  &=~(1 << READY);
		}
		
	}
	
	return 0;
}

// Find Counts per minute
uint16_t count_math(uint16_t newCount)
{
	static uint16_t old_cnt_array[10];		// Array used to store previous count values
	static uint8_t filterIndex;				// Used to run through array
	uint8_t i;
	uint16_t cntpermin;

	// Place counts during last second in array
	filterIndex++;
	if (filterIndex>=10) {
		filterIndex=0;
	}
	old_cnt_array[filterIndex] = newCount;
	
	// Sum up counts over latest 10 second period
	cntpermin = 0;
	for (i=0; i<10; i++) {
		cntpermin += old_cnt_array[i];
	}
	
	// Multiply with 6 to get CPM
	cntpermin *= 6;

	return cntpermin;
}

// Write value to LCD
void displayNumberOnLCDLine(uint16_t data, uint8_t line)
{
	char line_txt[8];

	// Format output
	utoa(data, line_txt, 10);

	// Write to LCD
	lcd_gotoxy(0, line);
	lcd_puts(line_txt);
}

// Parse and send a counts value
void sendCountValue(uint16_t count)
{
	char countString[7];	// uint16_t, so at most 5 characters, plus one termination char. Add one extra for good measure.
	
	// Place the datagram header in the string
	strcpy_P(stringBuffer, (PGM_P)pgm_read_word(&(send_protocol_string_table[4])));

	// Format output
	utoa(count, countString, 10);
	strcat(stringBuffer, countString);
	
	// Send the string
	sendSerialString(stringBuffer);
}

// Send the stored configuration over serial
void sendMyConfigurationData(void)
{
	// Get the correct strings from prog mem and send them
	for (unsigned char i = 0; i < 4; i++) {
		strcpy_P(stringBuffer, (PGM_P)pgm_read_word(&(send_protocol_string_table[i])));
		sendSerialString(stringBuffer);
	}
}

// Parse a character command
void parseCommandChar(char command) {
	switch (command) {
		case 'R':
			// Request for configuration received
			sendMyConfigurationData();
			break;
			
		case 'S':
			// Start transmitting command received
			flag_reg |= (1 << TX_ENABLED);
			break;
			
		case 'H':
			// Stop transmitting command received
			flag_reg &= ~(1 << TX_ENABLED);
			break;
	}
}
